import sys
import types
from inspect import getmembers, isfunction, isbuiltin, stack, getmodule
from functools import wraps
from importlib.machinery import PathFinder, SourceFileLoader
from zipimport import zipimporter
from nfer.nfer import report_interval, Interval, now
from pkgutil import iter_modules
import logging

log = logging.getLogger('nfer')

class watch(object):
    def __init__(self, clsname=None):
        self.clsname = clsname

    def __call__(self, func):
        self.func = func
        if self.clsname is not None:
            self.name = "%s.%s" % (self.clsname, self.func.__name__)
        else:
            self.name = self.func.__name__
        log.debug('Decorating %s' % self.name)

        def wrapped_func(*args, **kwargs):
            begin = now()
            result = self.func(*args, **kwargs)
            # have to create a transient object because update updates in place and doesn't return the modified dict
            passed_args = {'args':",".join(map(str, args))}
            kwargs_strs = dict(map(lambda d: (d[0], str(d[1])), kwargs.items() ))
            passed_args.update(kwargs_strs)
            report_interval(Interval(self.name, begin, now(), passed_args))
            return result
        return wrapped_func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        new_func = self.func.__get__(obj, type)
        return self.__class__(new_func)

symbol_cache = {}
PROBLEM_CLASSES = ["Process", "Thread"]

def instrument_module(module):
    global symbol_cache
    # iterate over the symbol names in the module
    for attr in dir(module):
        # get the symbol associated with the name
        sym = getattr(module, attr)

        # if the symbol is a type (class), instrument its methods
        if type(sym) == type.__class__:
            # skip the list of known problem classes
            if attr in PROBLEM_CLASSES:
                continue

            # iterate over the methods of the class
            for name, meth in getmembers(sym, predicate=isfunction):
                # skip any that we have seen before
                if f"{attr}.{name}" in symbol_cache:
                    continue
                # skip any built-ins, even if overrides - most are read-only anyway
                if isbuiltin(sym):
                    continue
                if name.startswith("__") and name.endswith("__"):
                    continue

                # overwrite the method with the wrapped method
                # print(f"overwrite {attr}.{name}")
                setattr(sym, name, watch(attr)(meth))
                # cache the class so we skip it next time
                symbol_cache[f"{attr}.{name}"] = sym

        # if the symbol is a function, instrument it
        elif callable(sym):
            # skip weird objects without __name__ attributes
            # there are lots of strange things in python modules and some of it is fragile
            if hasattr(sym, "__name__"):
                # skip any that we have seen before
                if attr in symbol_cache:
                    continue

                # overwrite the function in the module with the wrapped function
                # print(f"overwrite {attr}")
                setattr(module, attr, watch()(sym))
                # cache the function so we skip it next time
                symbol_cache[attr] = sym

# class InstrumentFinder:
#     packages=[]
#
#     @classmethod
#     def wrap_exec_module(cls, f):
#         @wraps(f)
#         def exec_module_wrapper(module):
#             ret = f(module)
#             instrument_module(module)
#             return ret
#
#         return exec_module_wrapper
#
#     @classmethod
#     def find_spec (cls, fullname, path=None, target=None):
#         """
#         Check PathFinder to see if it would load the file.
#         If so, wrap its exec_module function with our wrapper which will add instrumentation.
#         This is the least fragile way to do this, as far as I can tell.
#         """
#         path_spec = PathFinder.find_spec(fullname, path, target)
#         #print("Finding spec for fullname: ", fullname)
#         in_package = False
#         for package in InstrumentFinder.packages:
#             in_package = in_package or fullname.startswith(package)
#
#         print("checking %s %r %r" % (fullname, in_package, path_spec is not None))
#         if in_package and path_spec is not None:
#             print("and %r" % type(path_spec.loader))
#
#         if in_package and path_spec is not None:
#             if type(path_spec.loader) == SourceFileLoader or type(path_spec.loader) == zipimporter:
#                 print("instrumenting %s" % fullname)
#                 # unclear which is better - setattr or setting a new types.MethodType
#                 if hasattr(path_spec.loader, "exec_module"):
#                     setattr(path_spec.loader, "exec_module", InstrumentFinder.wrap_exec_module(path_spec.loader.exec_module))
#                     return path_spec
#                 # elif hasattr(path_spec.loader, "load_module"):
#                 #     setattr(path_spec.loader, "load_module", InstrumentFinder.wrap_exec_module(path_spec.loader.load_module))
#
#
#         return None
#
# def instrument_imports(package):
#     if not InstrumentFinder.packages:
#         # This puts our finder right in front of PathFinder
#         # This way we won't disrupt any built-in class loading, hopefully
#         pathfinder_index = sys.meta_path.index(PathFinder)
#         sys.meta_path.insert (pathfinder_index, InstrumentFinder)
#     InstrumentFinder.packages.append(package + ".")

def instrument_package(package):
    # passing a package name
    if isinstance(package, str):
        # check for the package
        if package not in sys.modules:
            # try to import it
            __import__(package)
        # then try getting it
        pkg = sys.modules[package]
    else:
        # otherwise we assume a module was passed
        pkg = package

    for importer, modname, ispkg in iter_modules(pkg.__path__):
        submod_name = "%s.%s" % (pkg.__name__, modname)
        if submod_name in sys.modules and not ispkg:
            instrument_module(sys.modules[submod_name])


def instrument(what=None):
    if what is None:
        stk = stack()
        try:
            instrument_module(sys.modules[getmodule(stk[1][0]).__name__])
        finally:
            # make sure we aren't keeping a reference to stack frames
            del stk
    else:
        # try loading it as a package
        instrument_package(what)


from uuid import uuid4
from enum import Enum

import _nfer

# requirements:
#    1. The user should be able to add rules whenever they like
#    2. The user should be able to use rules without naming the intervals
#    3. The user should be able to refer to named intervals by name not reference
#    4. The user should be able to run Python code when an interval is matched
#    5. The user should be able to add new events or intervals manually
#    6. The user should be able to specify conditions in Python syntax
#
#    def on_interval(interval):
#        if interval.id = 50:
#            print(interval.begin)
#        print(interval.end)
#        print(interval.id)
#        rule3 = when("boo").during(interval)
#        rule3.monitor(lambda i2: print(i2.begin))
#    rule = when("foo").before("bar").where("foo.id = bar.id | foo.id = 50").map("id", "foo.id")
#    rule = when("foo").before("bar").where("foo.id == bar.id or foo.id == 50").map("id", "foo.id")
#    rule = rule.name("baz")
#    rule2 = when("baz").meet("moo").where("baz.id = moo.id").map("id", "moo.id")
#    rule2.monitor(on_interval).monitor(lambda i3: print(i3.end))
#    Nfer.monitor_all()
#    Nfer.gui(port = 8080)


class Operator(Enum):
    atomic = 0
    before = 1
    meet = 2
    during = 3
    coincide = 4
    start = 5
    finish = 6
    overlap = 7
    slice = 8
    after = 9
    follow = 10
    contain = 11


def when(interval):
    # create a new Nfer instance
    # set the lhs to the passed interval, whether it's a string or a rule
    rule = AtomicRule(interval)
    return rule

class Rule:
    def __init__(self, result, lhs, operator, rhs, conditions, data, exclusion = False):
        """Constructs a new Nfer instance"""
        self.result = result
        self.lhs = lhs
        self.rhs = rhs
        self.operator = operator
        self.conditions = conditions
        self.data = data
        self.monitored = False
        self.exclusion = exclusion

    def where(self, condition):
        # add to the conditions
        self.conditions.append(condition)
        return self

    def map(self, key, value):
        # add to the map
        self.data[key] = value
        return self

    def name(self, value):
        self.result = value
        return self

    def to_rule_syntax(self):
        lhs = self.lhs
        rhs = self.rhs
        map = []
        map_str = ""
        where = []
        where_str = ""

        if isinstance(self.lhs, Rule):
            lhs = self.lhs.result

        if isinstance(self.rhs, Rule):
            rhs = self.rhs.result

        for key, value in self.data.items():
            map.append("{} -> {}".format(key, value))

        if map:
            map_str = " map {{ {} }}".format(", ".join(map))

        for condition in self.conditions:
            where.append(condition)

        if where:
            where_str = " where {}".format(" & ".join(where))

        if self.operator == Operator.atomic:
            return "{} :- {}{}{}".format(self.result, lhs, where_str, map_str)
        elif self.exclusion:
            return "{} :- {} unless {} {}{}{}".format(self.result, lhs, self.operator.name, rhs, where_str, map_str)
        else:
            return "{} :- {} {} {}{}{}".format(self.result, lhs, self.operator.name, rhs, where_str, map_str)


class AtomicRule(Rule):
    def __init__(self, lhs):
        """Constructs a new Nfer instance"""
        self.lhs = lhs
        self.rhs = None
        self.operator = Operator.atomic
        self.conditions = []
        self.data = {}
        self.result = "result{}".format(str(uuid4()))

    def before(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.before, interval, self.conditions, self.data)

    def meet(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.meet, interval, self.conditions, self.data)

    def during(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.during, interval, self.conditions, self.data)

    def coincide(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.coincide, interval, self.conditions, self.data)

    def start(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.start, interval, self.conditions, self.data)

    def finish(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.finish, interval, self.conditions, self.data)

    def overlap(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.overlap, interval, self.conditions, self.data)

    def slice(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.slice, interval, self.conditions, self.data)
    
    def unless_after(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.after, interval, self.conditions, self.data, exclusion=True)
    
    def unless_follow(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.follow, interval, self.conditions, self.data, exclusion=True)

    def unless_contain(self, interval):
        # set to the rhs to the passed interval, whether it's a string or a rule
        return Rule(self.result, self.lhs, Operator.contain, interval, self.conditions, self.data, exclusion=True)


import os
import pathlib
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.sdist import sdist
from distutils.dir_util import mkpath

# this directory
#PYDIR = os.path.dirname(os.path.realpath(pathlib.Path(__file__)))
# read in the Python module README markdown file
with open(os.path.join('python', 'README.md'), 'r') as readme:
    README = readme.read()

# the directory for the lexer/parser source
parser_dir = os.path.join('gensrc', 'parser')
# the source files for the lexer and parser
lexer_src = os.path.join(parser_dir, 'dsl.yy.c')
parser_src = os.path.join(parser_dir, 'dsl.tab.c')

nfer_c_extension = Extension('_nfer',
                    include_dirs = ['inc', parser_dir],
                    sources = [
                        os.path.join('src', 'types.c'),
                        os.path.join('src', 'log.c'),
                        os.path.join('src', 'nfer.c'),
                        os.path.join('src', 'dict.c'),
                        os.path.join('src', 'pool.c'),
                        os.path.join('src', 'file.c'),
                        os.path.join('src', 'map.c'),
                        os.path.join('src', 'stack.c'),
                        os.path.join('src', 'expression.c'),
                        os.path.join('src', 'ast.c'),
                        os.path.join('src', 'semantic.c'),
                        os.path.join('src', 'generate.c'),
                        os.path.join('src', 'astutil.c'),
                        os.path.join('src', 'memory.c'),
                        os.path.join('src', 'strings.c'),
                        os.path.join('src', 'learn.c'),
                        os.path.join('src', 'pyinterface.c'),
                        lexer_src, parser_src
                        ])

def generate_lexer_parser_source():
    "Checks for the lexer and parser source files and generates them if they are missing or outdated"

    # check if the parser source dir exists at all and create it if not
    if not os.path.exists(parser_dir):
        # using the distutils helper
        mkpath(parser_dir)

    # check for the lexer source and create it if it doesn't exist or is outdated
    lexer_l = os.path.join('src', 'dsl.l')
    lexer_src_is_old = True
    if os.path.exists(lexer_src) and os.path.exists(lexer_l):
        lexer_src_is_old = os.path.getmtime(lexer_src) < os.path.getmtime(lexer_l)
    
    if lexer_src_is_old:
        print('running flex')
        os.system('flex -o ' + lexer_src + ' ' + lexer_l)

    # check for the parser source and create it if it doesn't exist or is outdated
    parser_y = os.path.join('src', 'dsl.y')
    parser_src_is_old = True
    if os.path.exists(parser_src) and os.path.exists(parser_y):
        parser_src_is_old = os.path.getmtime(parser_src) < os.path.getmtime(parser_y)

    if parser_src_is_old:
        print('running bison')
        os.system('bison -d -o ' + parser_src + ' ' + parser_y)


class NferBuildPy(build_ext):
    def run(self):
        "overriding the build_ext command called by install to ensure the lexer/parser source exists"

        # honor the --dry-run flag
        if not self.dry_run:
            generate_lexer_parser_source()

        # call the super class method
        return build_ext.run(self)

class NferSourceDist(sdist):
    # you can override the behavior of any setuptools command this way
    def run(self):
        "Run preprocessor commands when building the source dist so installers don't need these dependencies."

        # honor the --dry-run flag
        if not self.dry_run:
            generate_lexer_parser_source()

        # call the super class method
        return sdist.run(self)


setup (name = 'NferModule',
       version = '0.9.10',
       description = 'The nfer language for Python',
       long_description = README,
       long_description_content_type = 'text/markdown',
       author = 'Sean Kauffman',
       author_email = 'skauffma@uwaterloo.ca',
       url = 'http://nfer.io/',
       package_dir = {'nfer': os.path.join('python', 'nfer')},
       packages = ['nfer'],
       ext_modules = [nfer_c_extension],
       cmdclass={'build_ext': NferBuildPy, 'sdist': NferSourceDist},
       install_requires=['aiohttp[speedups]', 'python-socketio>=5'],
       python_requires='>=3.5',
       include_package_data=True,
       zip_safe=False)

